# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import os.path

from .xloco import xLoco

from bs_lib import Pod


class Config:
    """
    Class that loads config values on demand from the database. The config values are cached
    for as long as the config is set to.
    """

    def __init__(self, pod: Pod, dao):
        """
        Constructor.

        :param pod: The pod to use.
        :param dao: The dao module to use.
        """
        self._pod       = pod
        self._dao       = dao
        self.cache_time = 60  # min
        self.cache      = {}


    def load(self, values: 'list|str', ignore_missing: bool = False, expand_env_vars: bool = True) -> dict:
        """
        Loads config values and caches them.

        :param values: List of string values to read.
        :param ignore_missing: If true, simply ignores missing parameters.
        :param expand_env_vars: If set to false, won't attempt to expand the env vars.
        :return: The config loaded.
        """
        fc = None

        if type(values) == str:
            values = [values]

        for val in values:
            if self.cache.get(val):
                continue

            if fc is None:
                fc = self._dao.dConfig(self._pod.dbcon)

            if fc.try_select_one_deft(val):
                if expand_env_vars:
                    self.cache[val] = os.path.expandvars(fc.rec.value)
                else:
                    self.cache[val] = fc.rec.value
            else:
                if not ignore_missing:
                    raise xLoco(f'Config setting not found [{val}]')

        return self.cache


    def read(self, value: str, expand_env_vars=True):
        """
        Reads a single value from the config.

        :param value: (string) The config value to read.
        :return: (string) The value.
        """
        val = self.cache.get(value)

        if val is None:
            self.load(value, False, expand_env_vars)

        return self.cache[value]



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#  Target Database = postgresql
#
import asyncio
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.snap_prov import tSnapProv
from bs_loco.db.tables.snap_prov_key import tSnapProvKey

from .snap_prov_lock_one import dSnapProvLockOne
from .snap_prov_select_one import dSnapProvSelectOne
from .snap_prov_update import dSnapProvUpdate
from .snap_prov_insert import dSnapProvInsert
from .snap_prov_delete_one import dSnapProvDeleteOne

class dSnapProv:

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self.rec    = tSnapProv()
        self._dbcon = dbcon

    def __enter__(self):
        """
        DAO enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        DAO exit.
        """
        pass

    def lock_one(self, _lock4update: mettle.db.DBLock, _expected: bool = True) -> bool:
        """
        Lock and select the record.

        :param _lock4update: The db lock object to use.
        :param _expected: If set to false, will not raise if record is not found.
        :return: True if selected, False if not found and _expected is set to False
        """
        _key = tSnapProvKey()

        self.rec._write_key(_key)

        return self.lock_one_by_key(_key, _lock4update, _expected)

    def lock_one_with(self, _rec: tSnapProv, _lock4update: mettle.db.DBLock, _expected: bool = True) -> bool:
        """
        Lock and select the record using a record.

        :param _rec: The table rec to select with and into.
        :param _lock4update: The db lock object to use.
        :param _expected: If set to false, will not raise if record is not found.
        :return: True if selected, False if not found and _expected is set to False
        """
        _key = tSnapProvKey()
        _rec._write_key(_key)

        self.lock_one_by_key(_key, _lock4update, _expected)
        _rec._copy_from(self.rec)
        return True

    def lock_one_deft(self,
                      valuedate: datetime.date,
                      corrtype_id: str,
                      corrprov_id: str,
                      _lock4update: mettle.db.DBLock,
                      _expected: bool = True) -> bool:
        """
        Attempt to lock and select the record by it's primary key columns.

        :param valuedate: datetime.date
        :param corrtype_id: str
        :param corrprov_id: str
        :param _lock4update:  The db lock object to use. 
        :param _expected: bool
        """
        _key = tSnapProvKey()

        _key.valuedate = valuedate
        _key.corrtype_id = corrtype_id
        _key.corrprov_id = corrprov_id

        return self.lock_one_by_key(_key, _lock4update, _expected)

    def lock_one_by_key(self, _key: tSnapProvKey, _lock4update: mettle.db.DBLock, _expected: bool = True) -> bool:
        """
        Attempt to lock and select the record with it's primary key.

        :param _key: Primary key of the table record.
        :param _lock4update: The db lock object to use.
        :param _expected: If set to false, will not raise if record is not found.
        :return: True if selected, False if not found and _expected is set to False
        """
        with dSnapProvLockOne(self._dbcon) as _qry:
            _dex = _lock4update.retrys - 1 if _lock4update.retrys > 0 else 0

            while _dex >= 0:
                try:
                    _qry.exec(_key)
                    if not _qry.fetch():
                        if not _expected:
                            return False

                        raise mettle.lib.xMettle("LockOne failed to return a record (dSnapProv)")

                    break
                except mettle.lib.xMettle as _x:
                    if _x.get_error_code() != mettle.lib.xMettle.eCode.DBLockNoWaitFailed:
                        raise _x

                    if _dex == 0:
                        raise mettle.lib.xMettle("Could not lock SnapProv record after %d attempt(s) at %d mili second intervals!" % (
_lock4update.retrys, _lock4update.mili_seconds), errCode = mettle.lib.xMettle.eCode.DBLockNoWaitFailed)

                    if _lock4update.mili_seconds > 0:
                        time.sleep(_lock4update.mili_seconds * 0.001)
                _dex -= 1

            self.rec = _qry.orec

            return True

    def try_select_one(self) -> bool:
        """
        Attempt to select the record.

        :return: True if the record was found.
        """
        _key = tSnapProvKey()

        self.rec._write_key(_key)

        return self.try_select_one_by_key(_key)

    def try_select_one_with(self, _rec: tSnapProv) -> bool:
        """
        Attempt to select the record using a table rec.

        :param _rec: The table rec to select with and into.
        :return: True if the row as selectd.
        """
        _key = tSnapProvKey()
        _rec._write_key(_key)

        if not self.try_select_one_by_key(_key):
            return False

        _rec._copy_from(self.rec)

        return True

    def try_select_one_deft(self,
                            valuedate: datetime.date,
                            corrtype_id: str,
                            corrprov_id: str) -> bool:
        """
        Attempt to select the record by it's primary key columns.

        :param valuedate: datetime.date
        :param corrtype_id: str
        :param corrprov_id: str
        :return: True if the row as selectd.
        """
        _key = tSnapProvKey()

        _key.valuedate = valuedate
        _key.corrtype_id = corrtype_id
        _key.corrprov_id = corrprov_id

        return self.try_select_one_by_key(_key)

    def try_select_one_by_key(self, _key: tSnapProvKey) -> bool:
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :returns: True if row was selected.
        """
        with dSnapProvSelectOne(self._dbcon) as _qry:
            _qry.exec(_key)
            if not _qry.fetch():
                return False

            self.rec = _qry.orec

        return True

    def select_one(self) -> "dSnapProv":
        """
        Selects the record, raises and error if it does not exist.

        :return: Self for convenience.
        """
        if not self.try_select_one():
            raise mettle.lib.xMettle("SelectOne failed to return a record (dSnapProv)")

        return self

    def select_one_with(self, _rec: tSnapProv) -> "dSnapProv":
        """
        Attempt to select the record with the table record.

        :param _rec: Table record.
        :returns: Self for convenience.
        """
        if not self.try_select_one_with(_rec):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dSnapProv)")

        return self

    def select_one_deft(self,
                        valuedate: datetime.date,
                        corrtype_id: str,
                        corrprov_id: str) -> "dSnapProv":
        """
        Selects the record by its primary key columns, raises and error if it does not exist.

        :param valuedate: datetime.date
        :param corrtype_id: str
        :param corrprov_id: str
        """
        _key = tSnapProvKey()

        _key.valuedate = valuedate
        _key.corrtype_id = corrtype_id
        _key.corrprov_id = corrprov_id

        if not self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dSnapProv)")

        return self

    def select_one_by_key(self, _key: tSnapProvKey) -> "dSnapProv":
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :return: Self for convenience.
        """
        if not self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dSnapProv)")

        return self

    def update_deft(self,
                    valuedate: datetime.date,
                    corrtype_id: str,
                    corrprov_id: str,
                    msg_sent: int,
                    msg_failed: int,
                    msg_retry: int,
                    avg_msg_time: float,
                    tot_msg_time: float):
        """
        Update the record by setting all column values.

        :param valuedate: datetime.date
        :param corrtype_id: str
        :param corrprov_id: str
        :param msg_sent: int
        :param msg_failed: int
        :param msg_retry: int
        :param avg_msg_time: float
        :param tot_msg_time: float
        """
        self.rec.valuedate = valuedate
        self.rec.corrtype_id = corrtype_id
        self.rec.corrprov_id = corrprov_id
        self.rec.msg_sent = msg_sent
        self.rec.msg_failed = msg_failed
        self.rec.msg_retry = msg_retry
        self.rec.avg_msg_time = avg_msg_time
        self.rec.tot_msg_time = tot_msg_time

        self.update()

    def update(self, irec: tSnapProv = None):
        """
        Update the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to update the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dSnapProvUpdate(self._dbcon) as _qry:
            _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)

    def insert_deft(self,
                    valuedate: datetime.date,
                    corrtype_id: str,
                    corrprov_id: str,
                    msg_sent: int,
                    msg_failed: int,
                    msg_retry: int,
                    avg_msg_time: float,
                    tot_msg_time: float):
        """
        Insert the record setting all column values.

        :param valuedate: datetime.date
        :param corrtype_id: str
        :param corrprov_id: str
        :param msg_sent: int
        :param msg_failed: int
        :param msg_retry: int
        :param avg_msg_time: float
        :param tot_msg_time: float
        """
        self.rec.valuedate = valuedate
        self.rec.corrtype_id = corrtype_id
        self.rec.corrprov_id = corrprov_id
        self.rec.msg_sent = msg_sent
        self.rec.msg_failed = msg_failed
        self.rec.msg_retry = msg_retry
        self.rec.avg_msg_time = avg_msg_time
        self.rec.tot_msg_time = tot_msg_time

        self.insert()

    def insert(self, irec: tSnapProv = None):
        """
        Insert the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to insert into the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dSnapProvInsert(self._dbcon) as _qry:
            _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)

    def delete_one_deft(self,
                        valuedate: datetime.date,
                        corrtype_id: str,
                        corrprov_id: str):
        """
        Delete one record by it's primary key columns.

        :param valuedate: datetime.date
        :param corrtype_id: str
        :param corrprov_id: str
        """
        _key = tSnapProvKey()

        _key.valuedate = valuedate
        _key.corrtype_id = corrtype_id
        _key.corrprov_id = corrprov_id

        self.delete_one(_key)

    def delete_one(self, irec: tSnapProvKey = None):
        """
        Delete the record by its primary key or irec if it is not None.

        :param irec: If not None use this record to delete record.
        """
        _key = tSnapProvKey()

        if irec:
            _key._copy_from(irec)
        else:
            self.rec._write_key(_key)

        with dSnapProvDeleteOne(self._dbcon) as _qry:
            _qry.exec(_key)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.snap_prov_key import tSnapProvKey
from bs_loco.db.tables.snap_prov import tSnapProv

class dSnapProvLockOne:

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tSnapProvKey()
        self.orec     = tSnapProv()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    def exec_deft(self,
                  valuedate: datetime.date,
                  corrtype_id: str,
                  corrprov_id: str) -> "dSnapProvLockOne":
        """
        Execute the query by setting all the inputs.

        :param valuedate: datetime.date
        :param corrtype_id: str
        :param corrprov_id: str
        :return: Self for convenience.
        """
        self.irec.valuedate = valuedate
        self.irec.corrtype_id = corrtype_id
        self.irec.corrprov_id = corrprov_id

        return self.exec()

    def exec(self, irec: tSnapProvKey = None) -> "dSnapProvLockOne":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("SnapProvLockOne", self._dbcon.STMNT_TYPE_READ)

        self._dbstmnt.sql("""select
  valuedate,
  corrtype_id,
  corrprov_id,
  msg_sent,
  msg_failed,
  msg_retry,
  avg_msgTime,
  tot_msgTime
from
  loco.SnapProv
where
  valuedate = :valuedate and
  corrtype_id = :corrtype_id and
  corrprov_id = :corrprov_id for update nowait""")

        self._dbstmnt.bind_in("valuedate", self.irec.valuedate, datetime.date)
        self._dbstmnt.bind_in("corrtype_id", self.irec.corrtype_id, str, 16)
        self._dbstmnt.bind_in("corrprov_id", self.irec.corrprov_id, str, 16)

        self._dbstmnt.bind_out("valuedate", datetime.date)
        self._dbstmnt.bind_out("corrtype_id", str, 16)
        self._dbstmnt.bind_out("corrprov_id", str, 16)
        self._dbstmnt.bind_out("msg_sent", int, 4)
        self._dbstmnt.bind_out("msg_failed", int, 4)
        self._dbstmnt.bind_out("msg_retry", int, 4)
        self._dbstmnt.bind_out("avg_msgTime", float)
        self._dbstmnt.bind_out("tot_msgTime", float)

        self._dbcon.lock(self._dbstmnt)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self

    def fetch(self) -> bool:
        """
        Fetches the next row from the executed query.

        :return: True if a row is retreived.
        """
        if not self._dbcon.fetch(self._dbstmnt):
            return False

        self.orec.valuedate = self._dbstmnt.result[0]
        self.orec.corrtype_id = self._dbstmnt.result[1]
        self.orec.corrprov_id = self._dbstmnt.result[2]
        self.orec.msg_sent = self._dbstmnt.result[3]
        self.orec.msg_failed = self._dbstmnt.result[4]
        self.orec.msg_retry = self._dbstmnt.result[5]
        self.orec.avg_msg_time = self._dbstmnt.result[6]
        self.orec.tot_msg_time = self._dbstmnt.result[7]

        return True

    def fetch_all(self, outList: tSnapProv.List, clearList: bool = True) -> int:
        """
        Fetch all the rows into the specified list and return the list count.

        :param outList:
        :param clearList:
        :return: Number of rows fetched.
        """
        if clearList:
            outList.clear()

        while self.fetch():
            outList.append(copy.deepcopy(self.orec))

        return len(outList)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.snap_prov import tSnapProv

class dSnapProvInsert:

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tSnapProv()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    async def exec_deft(self,
                        valuedate: datetime.date,
                        corrtype_id: str,
                        corrprov_id: str,
                        msg_sent: int,
                        msg_failed: int,
                        msg_retry: int,
                        avg_msg_time: float,
                        tot_msg_time: float) -> "dSnapProvInsert":
        """
        Execute the query by setting all the inputs.

        :param valuedate: datetime.date
        :param corrtype_id: str
        :param corrprov_id: str
        :param msg_sent: int
        :param msg_failed: int
        :param msg_retry: int
        :param avg_msg_time: float
        :param tot_msg_time: float
        :return: Self for convenience.
        """
        self.irec.valuedate = valuedate
        self.irec.corrtype_id = corrtype_id
        self.irec.corrprov_id = corrprov_id
        self.irec.msg_sent = msg_sent
        self.irec.msg_failed = msg_failed
        self.irec.msg_retry = msg_retry
        self.irec.avg_msg_time = avg_msg_time
        self.irec.tot_msg_time = tot_msg_time

        return await self.exec()

    async def exec(self, irec: tSnapProv = None) -> "dSnapProvInsert":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = await self._dbcon.statement("SnapProvInsert", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""insert into loco.SnapProv (
  valuedate,
  corrtype_id,
  corrprov_id,
  msg_sent,
  msg_failed,
  msg_retry,
  avg_msgTime,
  tot_msgTime
) values (
  :valuedate,
  :corrtype_id,
  :corrprov_id,
  :msg_sent,
  :msg_failed,
  :msg_retry,
  :avg_msgTime,
  :tot_msgTime
)""")

        self._dbstmnt.bind_in("valuedate", self.irec.valuedate, datetime.date)
        self._dbstmnt.bind_in("corrtype_id", self.irec.corrtype_id, str, 16)
        self._dbstmnt.bind_in("corrprov_id", self.irec.corrprov_id, str, 16)
        self._dbstmnt.bind_in("msg_sent", self.irec.msg_sent, int, 4)
        self._dbstmnt.bind_in("msg_failed", self.irec.msg_failed, int, 4)
        self._dbstmnt.bind_in("msg_retry", self.irec.msg_retry, int, 4)
        self._dbstmnt.bind_in("avg_msgTime", self.irec.avg_msg_time, float)
        self._dbstmnt.bind_in("tot_msgTime", self.irec.tot_msg_time, float)

        await self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import base64
import datetime
import json
import os.path
import ssl
import urllib
import urllib.parse
import urllib.request

from bs_lib  import Pod

from bs_loco.db.tables      import tMsg

from .provider_sms import ProviderSms

from bs_lib  import Common


class BulkSMSSms(ProviderSms):
    """
    The Bulk SMS provider
    """

    def __init__(self):
        """
        Constructor.
        """
        ProviderSms.__init__(self)


    def id(self) -> str:
        """
        Overload.
        """
        return 'bulksms'


    def initialize(self, pod: Pod, dao: object, cfg: dict):
        """
        Overload.
        """
        ProviderSms.initialize(self, pod, dao, cfg)

        Common.readDict(self._cfg, 'url',           str)
        Common.readDict(self._cfg, 'token',         str)
        Common.readDict(self._cfg, 'token-secret',  str)


    def destroy(self):
        """
        Overload.
        """
        self._client = None
        ProviderSms.destroy(self)


    def _provider_send(self, payload: dict, res_list: list, msg: tMsg, addr_list: list) -> tuple:
        """
        Overload.
        """
        try:
            dt_start = datetime.datetime.now()

            bs_tok = os.path.expandvars(self._cfg['token'])
            bs_sec = os.path.expandvars(self._cfg['token-secret'])

            for addr in addr_list:

                api_payload = {
                    "to"   : addr,
                    "body" : payload['msg'],
                }

                content = bytes(json.dumps(api_payload), 'utf8')
                auth    = base64.b64encode(bytes('%s:%s' % (bs_tok, bs_sec), 'utf8'))

                local_headers = {
                    'Content-Type'  : 'application/json',
                    'Authorization' : f'Basic {str(auth, "utf8")}',
                }

                ctx                = ssl.create_default_context()
                ctx.check_hostname = False
                ctx.verify_mode    = ssl.CERT_NONE

                conn = urllib.request.Request(
                    self._cfg['url'],
                    data    = content,
                    method  = 'POST',
                    headers = local_headers)

                try:
                    with urllib.request.urlopen(conn, timeout=30.0, context=ctx) as resp:
                        resp.read()

                except urllib.request.HTTPError as ex:
                    raise Exception(f'BulkSMS failed: [err:{ex.getcode()}, msg:{ex.read()}]')

            dt_end = datetime.datetime.now()
        except Exception as x:
            self._log.error(f'Exception caught sending sms [{msg}]')
            self._log.exception(x)
            return (False, None, str(x))

        return (True, dt_end - dt_start, None)



# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import datetime

from bs_lib import PodAsync

import bs_loco

from bs_loco       import util
from bs_loco.xloco import xLoco

from bs_loco.db.tables  import tMsg
from bs_loco.db.tables  import tMsgReq

from .loco_trig_async import LocoTrigAsync


class LocalDBAsync(LocoTrigAsync):
    """
    The local database trigger implementation.
    """

    def __init__(self):
        """
        Constructor.
        """
        LocoTrigAsync.__init__(self)
        self._dao = None


    def id(self) -> str:
        """
        Overload.
        """
        raise 'local_db'


    def name(self) -> str:
        """
        Overload.
        """
        'Local DB'


    async def initialize(self, pod: PodAsync, cfg: dict):
        """
        Overload.
        """
        await LocoTrigAsync.initialize(self, pod, cfg)

        if not cfg or not cfg.get('dao'):
            raise xLoco('Local DB requires a dao config value.')

        self._dao = bs_loco.dao_by_name(cfg.get('dao'), 'dao_async')


    def destroy(self):
        """
        Overload.
        """
        LocoTrigAsync.destroy(self)
        self._dao = None


    async def _trigger(self,
                       notype    : str,
                       corr_addr : dict,
                       key_pair  : dict = None,
                       res       : dict = None,
                       meta_data : dict = None) -> int:
        """
        Overload.
        """
        mreq     = self._dao.dMsgReq(self._pod.dbcon)
        mres     = None
        nt       = self._dao.dNoType(self._pod.dbcon)
        dtnow    = datetime.datetime.now()
        corrlist = []
        cnt      = 0

        await nt.select_one_deft(notype)

        if nt.rec.pref_corr:
            corrlist = nt.rec.pref_corr

            for corr in corr_addr:
                if corr not in corrlist:
                    corrlist.append(corr)

        else:
            corrlist = list(corr_addr.keys())

        for corr in corrlist:
            if nt.rec.corr_cnt and cnt >= nt.rec.corr_cnt:
                self._pod.log.info(f'Max correspondece count reached [{nt.rec.corr_cnt}], skipped: {corrlist[cnt:]}')
                break

            cnt += 1

        mreq.rec.clear()

        await mreq.insert_deft(
            notype,
            nt.rec.priority,
            tMsgReq.Status_Couplet.key_pending,
            dtnow,
            dtnow,
            datetime.datetime.min,
            corr_addr,
            key_pair,
            True if res and res.get('resources') else False,
            meta_data,
            self._pod.usr_id or '[Anonymous]')

        if res and res.get('resources'):
            for ritem in res['resources']:

                fnd = False

                for fc in ritem['for']:
                    if fc in corr:
                        fnd = True
                        break

                if not fnd:
                    continue

                if not mres:
                    mres = self._dao.dMsgReqRes(self._pod.dbcon)

                mres.rec.clear()

                if isinstance(ritem['content'], str):
                    content = bytes(ritem['content'], 'utf8')
                else:
                    content = ritem['content']

                await mres.insert_deft(
                    mreq.rec.id,
                    ritem['for'],
                    ritem['filename'],
                    ritem.get('mime') or util.mimetype_from_content(content),
                    content,
                    self._pod.usr_id)


        msg = self._dao.dMsg(self._pod.dbcon)

        for corr in corrlist:
            await msg.insert_deft(
                mreq.rec.id,
                corr,
                notype,
                mreq.rec.priority,
                tMsg.Status_Couplet.key_pending,
                0,
                0,
                datetime.datetime.min,
                0.0,
                '',
                '',
                '',
                self._pod.usr_id)

        return len(corrlist)



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#

import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

class oMsgStatusCntByReq(mettle.io.ISerializable):

    def __init__(self,
                 status: str = '',
                 cnt: int = 0):
        """
        Constructor.

        :param status: str
        :param cnt: int
        """
        self.status = status
        self.cnt = cnt

    def clear(self):
        """
        Clears all member variables for this database record.
        """
        self.status = ''
        self.cnt = 0

    def _name(self) -> str:
        """
        Name of the record.

        :return: The name.
        """
        return 'oMsgStatusCntByReq'

    def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
        """
        Serialize record to a stream.

        :param _w:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _w.write_start(_oname)
        _w.write_char("status", self.status)
        _w.write_int32("cnt", self.cnt)
        _w.write_end(_oname)

    def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
        """
        Deserialize record from a stream.

        :param _r:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _r.read_start(_oname)
        self.status = _r.read_char("status")
        self.cnt = _r.read_int32("cnt")
        _r.read_end(_oname)

    def _copy_from(self, rec: "oMsgStatusCntByReq"):
        """
        Copies the column from the rec into this record.

        :param rec: The source record.
        """
        self.status = rec.status
        self.cnt = rec.cnt

    def __repr__(self) -> str:
        return '<Msg [status:%.32s, cnt:%r]>' % (
            self.status, self.cnt)


    class List(list, mettle.io.ISerializable):
        """
        List Class
        """

        def _name(self) -> str:
            """
            Name of the list.

            :return: The name
            """
            return 'oMsgStatusCntByReq.List'

        def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
            """
            Serialize the list to a stream.

            :param _w:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _w.write_start_list(_oname, len(self))

            for _rec in self:
                _rec._serialize(_w)

            _w.write_end(_oname)

        def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
            """
            Deserialize the list from a stream.

            :param _r:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _cnt = _r.read_start_list(_oname)

            while _cnt >= 1:
                _rec  = oMsgStatusCntByReq()
                _cnt -= 1
                _rec._deserialize(_r)
                self.append(_rec)

            _r.read_end(_oname)

        def __repr__(self) -> str:
            return '<oMsgStatusCntByReq.List[%d]>' % (len(self))




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.msg import tMsg

class dMsgInsert:

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tMsg()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    async def exec_deft(self,
                        id: int,
                        msgreq_id: int,
                        corrtype_id: str,
                        notype_id: str,
                        priority: int,
                        status: str,
                        retry_cnt: int,
                        fail_cnt: int,
                        time_sent: datetime.datetime,
                        time_taken: float,
                        corrprov_id: str,
                        msglog: dict,
                        payload: dict,
                        modified_by: str,
                        tm_stamp: datetime.datetime) -> "dMsgInsert":
        """
        Execute the query by setting all the inputs.

        :param id: int
        :param msgreq_id: int
        :param corrtype_id: str
        :param notype_id: str
        :param priority: int
        :param status: str
        :param retry_cnt: int
        :param fail_cnt: int
        :param time_sent: datetime.datetime
        :param time_taken: float
        :param corrprov_id: str
        :param msglog: dict
        :param payload: dict
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        :return: Self for convenience.
        """
        self.irec.id = id
        self.irec.msgreq_id = msgreq_id
        self.irec.corrtype_id = corrtype_id
        self.irec.notype_id = notype_id
        self.irec.priority = priority
        self.irec.status = status
        self.irec.retry_cnt = retry_cnt
        self.irec.fail_cnt = fail_cnt
        self.irec.time_sent = time_sent
        self.irec.time_taken = time_taken
        self.irec.corrprov_id = corrprov_id
        self.irec.msglog = msglog
        self.irec.payload = payload
        self.irec.modified_by = modified_by
        self.irec.tm_stamp = tm_stamp

        return await self.exec()

    async def exec(self, irec: tMsg = None) -> "dMsgInsert":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = await self._dbcon.statement("MsgInsert", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""insert into loco.Msg (
  msgreq_id,
  corrtype_id,
  notype_id,
  priority,
  status,
  retry_cnt,
  fail_cnt,
  time_sent,
  time_taken,
  corrprov_id,
  msglog,
  payload,
  modified_by,
  tm_stamp
) values (
  :msgreq_id,
  :corrtype_id,
  :notype_id,
  :priority,
  :status,
  :retry_cnt,
  :fail_cnt,
  :time_sent,
  :time_taken,
  :corrprov_id,
  :msglog,
  :payload,
  :modified_by,
  :tm_stamp
)""")

        self.irec.tm_stamp = await self._dbcon.get_timestamp()

        self._dbstmnt.bind_in("msgreq_id", self.irec.msgreq_id, int, 8)
        self._dbstmnt.bind_in("corrtype_id", self.irec.corrtype_id, str, 16)
        self._dbstmnt.bind_in("notype_id", self.irec.notype_id, str, 128)
        self._dbstmnt.bind_in("priority", self.irec.priority, int, 4)
        self._dbstmnt.bind_in("status", self.irec.status, str, 1)
        self._dbstmnt.bind_in("retry_cnt", self.irec.retry_cnt, int, 4)
        self._dbstmnt.bind_in("fail_cnt", self.irec.fail_cnt, int, 4)
        self._dbstmnt.bind_in("time_sent", self.irec.time_sent, datetime.datetime)
        self._dbstmnt.bind_in("time_taken", self.irec.time_taken, float)
        self._dbstmnt.bind_in("corrprov_id", self.irec.corrprov_id, str, 16)
        self._dbstmnt.bind_in("msglog", self.irec.msglog, dict)
        self._dbstmnt.bind_in("payload", self.irec.payload, dict)
        self._dbstmnt.bind_in("modified_by", self.irec.modified_by, str, 128)
        self._dbstmnt.bind_in("tm_stamp", self.irec.tm_stamp, datetime.datetime)

        await self._dbcon.execute(self._dbstmnt)

        self.irec.id = await self._dbcon.post_get_sequence(self._dbstmnt, "loco.Msg", "id", 8)

        if irec:
            irec._copy_from(self.irec)

        return self




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
from bs_lib import PodAsync

from .base_loco_trig import BaseLocoTrig


class LocoTrigAsync(BaseLocoTrig):
    """
    This is the base class for firing off async loco message requests
    into the system.
    """

    def __init__(self):
        """
        Constructor.
        """
        BaseLocoTrig.__init__(self)


    def __del__(self):
        """
        Destructor.
        """
        self.destroy()


    async def initialize(self, pod: PodAsync, cfg: dict):
        """
        Virtual method to initialize the loco trigger interface.

        :param pod: The pod to use.
        :param cfg: Any configs needed for the trigger, can be None.
        """
        self.destroy()
        self._pod = pod
        self._cfg = cfg


    def destroy(self):
        """
        Virtual method to destroy the content management object.
        """
        self._pod = None
        self._cfg = None


    async def trig(self,
                   notype    : str,
                   corr_addr : dict,
                   key_pair  : dict = None,
                   res       : dict = None,
                   meta_data : dict = None) -> int:
        """
        Wrapper method that validates and then triggers a new correspondence message to correspondence addresses.

        :param notype: The notification type identifier.
        :param corr_addr: The addresses for each correspondence type required.
        :param key_pair: Additional data dictionary to use for messsage subsitution.
        :param res: And additional file resources to add to the out going correspondence.
        :param meta_data: Optional meta data.
        :return: The number of correspondence messages queued.
        """
        self._validate_dict(self.SCHEMA_ADDR, corr_addr, 'corr_addr', True)
        self._validate_dict(self.SCHEMA_RESOURCE, res, 'res', False)

        return await self._trigger(notype, corr_addr, key_pair, res, meta_data)


    async def _trigger(self,
                       notype    : str,
                       corr_addr : dict,
                       key_pair  : dict = None,
                       res       : dict = None,
                       meta_data : dict = None) -> int:
        """
        Pure virtual method that triggers the actual correspondence messages.

        :param notype: The notification type identifier.
        :param corr_addr: The addresses for each correspondence type required.
        :param key_pair: Additional data dictionary to use for messsage subsitution.
        :param res: And additional file resources to add to the out going correspondence.
        :param meta_data: Optional meta data for the trigger.
        :return: The number of correspondence messages queued.
        """
        raise Exception('_trigger() not implemented.')



# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import boto3
import datetime
import os.path

from bs_lib   import Pod
from bs_lib   import common

from bs_loco.db.tables import tMsg

from .provider_sms import ProviderSms


class AwsSnsSms(ProviderSms):
    """
    The AWS Simple Notification Service provider
    """

    def __init__(self):
        """
        Constructor.
        """
        ProviderSms.__init__(self)
        self._client = None


    def id(self) -> str:
        """
        Overload.
        """
        return 'awssns'


    def initialize(self, pod: Pod, dao: object, cfg: dict):
        """
        Overload.
        """
        ProviderSms.initialize(self, pod, dao, cfg)

        common.read_dict(self._cfg, 'aws_access_key_id',      str)
        common.read_dict(self._cfg, 'aws_secret_access_key',  str)
        common.read_dict(self._cfg, 'aws_region',             str)

        self._client = boto3.client("sns",
                                    aws_access_key_id     = os.path.expandvars(cfg['aws_access_key_id']),
                                    aws_secret_access_key = os.path.expandvars(cfg['aws_secret_access_key']),
                                    region_name           = os.path.expandvars(cfg['aws_region']))


    def destroy(self):
        """
        Overload.
        """
        self._client = None
        ProviderSms.destroy(self)


    def _provider_send(self, payload: dict, res_list: list, msg: tMsg, addr_list: list) -> tuple:
        """
        Overload.
        """
        try:
            dt_start = datetime.datetime.now()

            for addr in addr_list:
                self._client.publish(PhoneNumber=addr, Message=payload['msg'])

            dt_end = datetime.datetime.now()
        except Exception as x:
            self._log.error(f'Exception caught sending sms [{msg}]')
            self._log.exception(x)
            return (False, None, str(x))

        return (True, dt_end - dt_start, None)



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#  Target Database = postgresql
#
from .config import dConfig
from .config_select_one import dConfigSelectOne
from .config_lock_one import dConfigLockOne
from .config_update import dConfigUpdate
from .config_insert import dConfigInsert
from .config_search import dConfigSearch
from .corr_prov import dCorrProv
from .corr_prov_select_one import dCorrProvSelectOne
from .corr_prov_lock_one import dCorrProvLockOne
from .corr_prov_update import dCorrProvUpdate
from .corr_prov_delete_one import dCorrProvDeleteOne
from .corr_prov_insert import dCorrProvInsert
from .corr_prov_for_service import dCorrProvForService
from .corr_prov_cfg import dCorrProvCfg
from .corr_prov_cfg_select_one import dCorrProvCfgSelectOne
from .corr_prov_cfg_lock_one import dCorrProvCfgLockOne
from .corr_prov_cfg_update import dCorrProvCfgUpdate
from .corr_prov_cfg_delete_one import dCorrProvCfgDeleteOne
from .corr_prov_cfg_insert import dCorrProvCfgInsert
from .corr_type import dCorrType
from .corr_type_select_one import dCorrTypeSelectOne
from .corr_type_lock_one import dCorrTypeLockOne
from .corr_type_update import dCorrTypeUpdate
from .corr_type_insert import dCorrTypeInsert
from .corr_type_select_all import dCorrTypeSelectAll
from .corr_type_allow import dCorrTypeAllow
from .corr_type_allow_select_one import dCorrTypeAllowSelectOne
from .corr_type_allow_lock_one import dCorrTypeAllowLockOne
from .corr_type_allow_update import dCorrTypeAllowUpdate
from .corr_type_allow_delete_one import dCorrTypeAllowDeleteOne
from .corr_type_allow_insert import dCorrTypeAllowInsert
from .corr_type_allow_regex_for_type import dCorrTypeAllowRegexForType
from .msg import dMsg
from .msg_select_one import dMsgSelectOne
from .msg_lock_one import dMsgLockOne
from .msg_update import dMsgUpdate
from .msg_delete_one import dMsgDeleteOne
from .msg_insert import dMsgInsert
from .msg_by_msg_req import dMsgByMsgReq
from .msg_for_processing import dMsgForProcessing
from .msg_status_cnt_by_req import dMsgStatusCntByReq
from .msg_hist import dMsgHist
from .msg_hist_select_one import dMsgHistSelectOne
from .msg_hist_update import dMsgHistUpdate
from .msg_hist_insert import dMsgHistInsert
from .msg_hist_search import dMsgHistSearch
from .msg_req import dMsgReq
from .msg_req_select_one import dMsgReqSelectOne
from .msg_req_lock_one import dMsgReqLockOne
from .msg_req_update import dMsgReqUpdate
from .msg_req_delete_one import dMsgReqDeleteOne
from .msg_req_insert import dMsgReqInsert
from .msg_req_update_status import dMsgReqUpdateStatus
from .msg_req_pending_changed import dMsgReqPendingChanged
from .msg_req_busy_changed import dMsgReqBusyChanged
from .msg_req_for_archive import dMsgReqForArchive
from .msg_req_res import dMsgReqRes
from .msg_req_res_select_one import dMsgReqResSelectOne
from .msg_req_res_lock_one import dMsgReqResLockOne
from .msg_req_res_update import dMsgReqResUpdate
from .msg_req_res_delete_one import dMsgReqResDeleteOne
from .msg_req_res_insert import dMsgReqResInsert
from .msg_req_res_by_msg_req import dMsgReqResByMsgReq
from .no_dict import dNoDict
from .no_dict_select_one import dNoDictSelectOne
from .no_dict_lock_one import dNoDictLockOne
from .no_dict_update import dNoDictUpdate
from .no_dict_delete_one import dNoDictDeleteOne
from .no_dict_insert import dNoDictInsert
from .no_dict_search import dNoDictSearch
from .no_res import dNoRes
from .no_res_select_one import dNoResSelectOne
from .no_res_lock_one import dNoResLockOne
from .no_res_update import dNoResUpdate
from .no_res_delete_one import dNoResDeleteOne
from .no_res_insert import dNoResInsert
from .no_res_search import dNoResSearch
from .no_templ import dNoTempl
from .no_templ_select_one import dNoTemplSelectOne
from .no_templ_lock_one import dNoTemplLockOne
from .no_templ_update import dNoTemplUpdate
from .no_templ_insert import dNoTemplInsert
from .no_templ_by_no_type import dNoTemplByNoType
from .no_templ_cfg import dNoTemplCfg
from .no_templ_cfg_select_one import dNoTemplCfgSelectOne
from .no_templ_cfg_lock_one import dNoTemplCfgLockOne
from .no_templ_cfg_update import dNoTemplCfgUpdate
from .no_templ_cfg_insert import dNoTemplCfgInsert
from .no_type import dNoType
from .no_type_select_one import dNoTypeSelectOne
from .no_type_lock_one import dNoTypeLockOne
from .no_type_update import dNoTypeUpdate
from .no_type_delete_one import dNoTypeDeleteOne
from .no_type_insert import dNoTypeInsert
from .snap_msg_req import dSnapMsgReq
from .snap_msg_req_select_one import dSnapMsgReqSelectOne
from .snap_msg_req_lock_one import dSnapMsgReqLockOne
from .snap_msg_req_update import dSnapMsgReqUpdate
from .snap_msg_req_delete_one import dSnapMsgReqDeleteOne
from .snap_msg_req_insert import dSnapMsgReqInsert
from .snap_prov import dSnapProv
from .snap_prov_select_one import dSnapProvSelectOne
from .snap_prov_lock_one import dSnapProvLockOne
from .snap_prov_update import dSnapProvUpdate
from .snap_prov_delete_one import dSnapProvDeleteOne
from .snap_prov_insert import dSnapProvInsert



# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import os

from bs_lib import Pod

from bs_loco.xloco     import xLoco
from bs_loco           import constants

from .provider import Provider

# The below list is formatter as "<corrtype_id>-<provider_id>-<module_path>"

LOCO_PROVIDER_LIST = [
    'email-smtp-bs_loco.corrprov.smtp_email.SmtpEmail',
    'email-logfile-bs_loco.corrprov.logfile_email.LogfileEmail',
    'sms-awssns-bs_loco.corrprov.awssns_sms.AwsSnsSms',
    'sms-bulksms-bs_loco.corrprov.bulksms_sms.BulkSMSSms',
    'sms-logfile-bs_loco.corrprov.logfile_sms.LogfileSms',
]


if os.environ.get('BITSMITHS_LOCO_CUST_PROVIDERS'):
    for cprov in os.environ.get('METTLE_MKFILE_CUST_GENES').split(os.path.pathsep):
        LOCO_PROVIDER_LIST.append(cprov)

_providers = {
    constants.CORRTYPE_EMAIL : {},
    constants.CORRTYPE_SMS   : {},
}

for prov in LOCO_PROVIDER_LIST:
    parts = prov.split('-')

    if len(parts) != 3:
        raise Exception(f'Provider Item [{prov}] expected to be in in format: <corrtype_id>-<provider_id>-<module_path>')

    if parts[0] not in _providers.keys():
        raise Exception(f'Provider Item [{parts[0]}] is not an expected correspondence type. Expected: [{_providers.keys()}]')

    _providers[parts[0]][parts[1]] = { 'path': parts[2], 'prov': None }


del LOCO_PROVIDER_LIST



def get_provider(pod: Pod, dao, corrtype_id: str) -> Provider:
    """
    Factory to load a correspondence provider by id.

    :param pod: The pod to use.
    :param dao: The dao module to use.
    :param corrtype_id: The correspondence type to load.
    """
    from bs_lib import common

    plist = []

    if not dao.dCorrProvForService(pod.dbcon).exec_deft(corrtype_id).fetch_all(plist):
        raise xLoco('No active providers configured! [%s] .' % corrtype_id)

    if len(plist) > 1:
        pod.log.warning('Multiple active providers detected. Selecting the most recently chagned.')

        pod.log.warning(f'  - selected: {plist[0]}')

        for x in plist[1:]:
            pod.log.warning(f'  - ignored : {x}')

    rec    = plist[0]
    res    = None
    ct_obj = _providers.get(rec.corrtype_id)

    if ct_obj:
        prov_obj = ct_obj.get(rec.corrprov_id)

        if prov_obj:
            if not prov_obj['prov']:
                prov, exc    = common.import_dyn_pluggin(prov_obj['path'], False, True)
                prov_obj['prov'] = prov

            res = prov_obj['prov']()

    if not res:
        raise xLoco(f'Provider not found [{rec}].')

    if res.id() != rec.corrprov_id:
        raise Exception(f'Factory error, database corrprov_id does not match object corrprov_id: {rec}')

    if res.corrtype_id() != rec.corrtype_id:
        raise Exception(f'Factory error, database corrtype_id does not match object corrtype_id: {rec}')

    res.initialize(pod, dao, rec.cfg)

    return res


__all__ = [ 'get_provider' ]



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.msg_key import tMsgKey
from bs_loco.db.tables.msg import tMsg

class dMsgSelectOne:

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tMsgKey()
        self.orec     = tMsg()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    async def exec_deft(self,
                        id: int) -> "dMsgSelectOne":
        """
        Execute the query by setting all the inputs.

        :param id: int
        :return: Self for convenience.
        """
        self.irec.id = id

        return await self.exec()

    async def exec(self, irec: tMsgKey = None) -> "dMsgSelectOne":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = await self._dbcon.statement("MsgSelectOne", self._dbcon.STMNT_TYPE_READ)

        self._dbstmnt.sql("""select
  id,
  msgreq_id,
  corrtype_id,
  notype_id,
  priority,
  status,
  retry_cnt,
  fail_cnt,
  time_sent,
  time_taken,
  corrprov_id,
  msglog,
  payload,
  modified_by,
  tm_stamp
from
  loco.Msg
where
  id = :id""")

        self._dbstmnt.bind_in("id", self.irec.id, int, 8)

        self._dbstmnt.bind_out("id", int, 8)
        self._dbstmnt.bind_out("msgreq_id", int, 8)
        self._dbstmnt.bind_out("corrtype_id", str, 16)
        self._dbstmnt.bind_out("notype_id", str, 128)
        self._dbstmnt.bind_out("priority", int, 4)
        self._dbstmnt.bind_out("status", str, 1)
        self._dbstmnt.bind_out("retry_cnt", int, 4)
        self._dbstmnt.bind_out("fail_cnt", int, 4)
        self._dbstmnt.bind_out("time_sent", datetime.datetime)
        self._dbstmnt.bind_out("time_taken", float)
        self._dbstmnt.bind_out("corrprov_id", str, 16)
        self._dbstmnt.bind_out("msglog", dict)
        self._dbstmnt.bind_out("payload", dict)
        self._dbstmnt.bind_out("modified_by", str, 128)
        self._dbstmnt.bind_out("tm_stamp", datetime.datetime)

        await self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self

    async def fetch(self) -> bool:
        """
        Fetches the next row from the executed query.

        :return: True if a row is retreived.
        """
        if not await self._dbcon.fetch(self._dbstmnt):
            return False

        self.orec.id = self._dbstmnt.result[0]
        self.orec.msgreq_id = self._dbstmnt.result[1]
        self.orec.corrtype_id = self._dbstmnt.result[2]
        self.orec.notype_id = self._dbstmnt.result[3]
        self.orec.priority = self._dbstmnt.result[4]
        self.orec.status = self._dbstmnt.result[5]
        self.orec.retry_cnt = self._dbstmnt.result[6]
        self.orec.fail_cnt = self._dbstmnt.result[7]
        self.orec.time_sent = self._dbstmnt.result[8]
        self.orec.time_taken = self._dbstmnt.result[9]
        self.orec.corrprov_id = self._dbstmnt.result[10]
        self.orec.msglog = self._dbstmnt.result[11]
        self.orec.payload = self._dbstmnt.result[12]
        self.orec.modified_by = self._dbstmnt.result[13]
        self.orec.tm_stamp = self._dbstmnt.result[14]

        return True

    async def fetch_all(self, outList: tMsg.List, clearList: bool = True) -> int:
        """
        Fetch all the rows into the specified list and return the list count.

        :param outList:
        :param clearList:
        :return: Number of rows fetched.
        """
        if clearList:
            outList.clear()

        while await self.fetch():
            outList.append(copy.deepcopy(self.orec))

        return len(outList)




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
from bs_loco       import constants
from bs_loco.xloco import xLoco

from bs_loco.db.tables   import tMsgReq

from .provider import Provider


class ProviderEmail(Provider):
    """
    This is the base email provder overload.
    """

    def __init__(self):
        """
        Constructor.
        """
        Provider.__init__(self)


    def corrtype_id(self) -> str:
        """
        Overload.
        """
        return constants.CORRTYPE_EMAIL


    def _build_payload(self, req: tMsgReq) -> dict:
        """
        Overload.
        """
        templ = self._cache['templ'][req.notype_id]
        kpair = self._build_msg_keypairs(req)

        return {
            'subject'   : self._template_subst(templ.get('subject'),   kpair),
            'body-html' : self._template_subst(templ.get('body-html'), kpair),
            'body-text' : self._template_subst(templ.get('body-text'), kpair),
        }


    def _validate_template(self, templ: dict) -> dict:
        """
        Overload.
        """
        for val in [ 'subject', 'body-html', 'body-text' ]:
            obj = templ.get(val)

            if not obj:
                raise xLoco(f'Invalid email template, [{val}] missing or empty [corrprov:{self.id()}'
                            f', corrtype:{self.corrtype_id()}, templ:{templ}]')

        return templ



# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #

import datetime
import json

from mettle.io.py_json_dict_writer import PyJsonDictWriter

from bs_lib import Pod

import bs_loco

from bs_loco.db.tables   import tMsg
from bs_loco.db.tables   import tMsgReq
from bs_loco.db.tables   import tMsgReqRes


class BatchArchiveMsgs:
    """
    Batch object to archive messages.
    """

    def __init__(self, pod: Pod, dao_name: str):
        """
        Constructor

        :param pod: The pod to use.
        :param dao_name: The dao objects to use.
        """
        self._pod = pod
        self._log = pod.log
        self._dao = bs_loco.dao_by_name(dao_name)


    def run(self, dtnow: datetime.datetime) -> None:
        """
        Main run method.

        :param dtnow: The time of this archive run
        """
        arch_date   = datetime.datetime(dtnow.year, dtnow.month, dtnow.day, 0, 0, 0) - datetime.timedelta(days=1)
        arch_qry    = self._dao.dMsgReqForArchive(self._pod.dbcon)
        msg_qry     = self._dao.dMsgByMsgReq(self._pod.dbcon)
        msgattr_qry = self._dao.dMsgReqResByMsgReq(self._pod.dbcon)
        cnt         = 0

        msg_hist  = self._dao.dMsgHist(self._pod.dbcon)
        msg       = self._dao.dMsg(self._pod.dbcon)
        msg_req   = self._dao.dMsgReq(self._pod.dbcon)
        msg_res   = self._dao.dMsgReqRes(self._pod.dbcon)

        list_msgs = tMsg.List()
        list_res  = tMsgReqRes.List()
        snap_req  = {}
        snap_prov = {}
        json_obj  = {}

        self._log.info(f'Archiving messages from: {arch_date}')

        arch_qry.exec_deft(arch_date)

        while arch_qry.fetch():
            cnt += 1
            self._log.debug(f' - archiving [cnt:{cnt}, msg_req:{arch_qry.orec}]')

            list_msgs.clear()
            list_res.clear()
            json_obj.clear()

            msg_qry.exec_deft(arch_qry.orec.id).fetch_all(list_msgs)
            msgattr_qry.exec_deft(arch_qry.orec.id, '').fetch_all(list_res)

            writer = PyJsonDictWriter(json_obj)

            arch_qry.orec._serialize(writer)
            list_msgs._serialize(writer)
            list_res._serialize(writer)

            self._update_snaps(arch_qry.orec, list_msgs, snap_req, snap_prov)

            msg_hist.insert_deft(arch_qry.orec.id,
                                 arch_qry.orec.status,
                                 arch_qry.orec.time_to_send,
                                 json.dumps(json_obj))

            for a in list_res:
                msg_res.delete_one_deft(a.id)

            for m in list_msgs:
                msg.delete_one_deft(m.id)

            msg_req.delete_one_deft(arch_qry.orec.id)

            if cnt % 10000 == 0:
                self._log.info(f' - archived [{cnt}] messages.')
                self._commit_changes(snap_req, snap_prov)

        if cnt > 0 and not cnt % 10000 == 0:
            self._commit_changes(snap_req, snap_prov)

        self._log.info(f'Total messages archived [{cnt}]')


    def _commit_changes(self, snap_req: dict, snap_prov: dict):
        """
        Commits the current changes and continues

        :param snap_req: snapshots on msg requests
        :param snap_prov: snapshots on providers
        """
        smr = self._dao.dSnapMsgReq(self._pod.dbcon)
        sp  = self._dao.dSnapProv(self._pod.dbcon)

        for key, item in snap_req.items():
            smr.update(item)

        for key, item in snap_prov.items():
            item.avg_msg_time = item.tot_msg_time / item.msg_sent
            sp.update(item)

        self._pod.dbcon.commit()


    def _update_snaps(self, req_rec: tMsgReq, list_msgs: list, snap_req: dict, snap_prov: dict):
        """
        Update the statistic tables

        :param req_rec: message request rec
        :param list_msgs: messages list
        :param snap_req: snapshots on msg requests
        :param snap_prov: snapshots on providers
        """
        key = '%s.%s' % (req_rec.time_to_send.strftime('%Y%m%d'), req_rec.notype_id)

        if key not in snap_req.keys():
            smr = self._dao.dSnapMsgReq(self._pod.dbcon)

            if smr.try_select_one_deft(req_rec.time_to_send.date(), req_rec.notype_id):
                snap_req[key] = smr.rec
            else:
                smr.rec.valuedate = req_rec.time_to_send.date()
                smr.rec.notype_id = req_rec.notype_id

                smr.insert()

                snap_req[key] = smr.rec

        rec          = snap_req[key]
        rec.tot_req += 1

        if req_rec.status == tMsgReq.Status_Couplet.key_processed:
            rec.tot_proc += 1
        elif req_rec.status == tMsgReq.Status_Couplet.key_cancelled:
            rec.tot_cancelled += 1
        elif req_rec.status == tMsgReq.Status_Couplet.key_partially_cancelled:
            rec.tot_part_canc  += 1

        rec.tot_msgs += len(list_msgs)

        for msgitem in list_msgs:
            if msgitem.status == tMsg.Status_Couplet.key_sent:
                rec.tot_msgs_sent += 1
            else:
                rec.tot_msgs_failed += 1

            rec.tot_msgs_retrys += msgitem.retry_cnt

        # do provider snapshots
        for msgitem in list_msgs:
            if msgitem.corrprov_id == '':
                continue

            key = '%s.%s.%s' % (req_rec.time_to_send.strftime('%Y%m%d'), msgitem.corrtype_id, msgitem.corrprov_id)

            if key not in snap_prov.keys():
                sp = self._dao.dSnapProv(self._pod.dbcon)

                if sp.try_select_one_deft(req_rec.time_to_send.date(), msgitem.corrtype_id, msgitem.corrprov_id):
                    snap_prov[key] = sp.rec
                else:
                    sp.rec.valuedate    = req_rec.time_to_send.date()
                    sp.rec.corrprov_id  = msgitem.corrprov_id
                    sp.rec.corrtype_id  = msgitem.corrtype_id

                    sp.insert()

                    snap_prov[key] = sp.rec

            rec = snap_prov[key]

            if msgitem.status == tMsg.Status_Couplet.key_sent:
                rec.msg_sent     += 1
                rec.tot_msg_time += msgitem.time_taken
            else:
                rec.msg_failed += 1

            rec.msg_retry += msgitem.retry_cnt



from setuptools import setup

import shutil
import os.path

lpath = os.path.expandvars("$LOCAL_PATH")

shutil.copyfile(os.path.join(lpath, 'LICENCE'), os.path.join(lpath, 'products', 'loco', 'python3', 'LICENCE'))


def build_sql_files():
    import os

    ext_order = ['.table', '.constraint', '.index']

    for (root, dirs, files) in os.walk('../mettledb/sqldef'):
        if not files or dirs:
            continue

        sql_name  = os.path.split(root)[-1]
        dest_file = os.path.join(f'bs_loco/sql/{sql_name}.sql')

        os.system(f'mettle-sql-build --clean {dest_file}')

        for ext in ext_order:
            for fl in files:
                if fl.endswith(ext):
                    os.system(f'mettle-sql-build --file {os.path.join(root, fl)} {dest_file}')


        init_file = os.path.join('..', 'sql', f'initdb.{sql_name}.sql')

        if os.path.exists(init_file):
            os.system(f'mettle-sql-build --file {init_file} {dest_file}')


build_sql_files()

setup()



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.corr_type_allow import tCorrTypeAllow

class dCorrTypeAllowInsert:

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tCorrTypeAllow()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    def exec_deft(self,
                  corrtype_id: str,
                  allow_id: str,
                  status: str,
                  regex: str,
                  modified_by: str,
                  tm_stamp: datetime.datetime) -> "dCorrTypeAllowInsert":
        """
        Execute the query by setting all the inputs.

        :param corrtype_id: str
        :param allow_id: str
        :param status: str
        :param regex: str
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        :return: Self for convenience.
        """
        self.irec.corrtype_id = corrtype_id
        self.irec.allow_id = allow_id
        self.irec.status = status
        self.irec.regex = regex
        self.irec.modified_by = modified_by
        self.irec.tm_stamp = tm_stamp

        return self.exec()

    def exec(self, irec: tCorrTypeAllow = None) -> "dCorrTypeAllowInsert":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("CorrTypeAllowInsert", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""insert into loco.CorrTypeAllow (
  corrtype_id,
  allow_id,
  status,
  regex,
  modified_by,
  tm_stamp
) values (
  :corrtype_id,
  :allow_id,
  :status,
  :regex,
  :modified_by,
  :tm_stamp
)""")

        self.irec.tm_stamp = self._dbcon.get_timestamp()

        self._dbstmnt.bind_in("corrtype_id", self.irec.corrtype_id, str, 16)
        self._dbstmnt.bind_in("allow_id", self.irec.allow_id, str, 32)
        self._dbstmnt.bind_in("status", self.irec.status, str, 1)
        self._dbstmnt.bind_in("regex", self.irec.regex, str, 256)
        self._dbstmnt.bind_in("modified_by", self.irec.modified_by, str, 128)
        self._dbstmnt.bind_in("tm_stamp", self.irec.tm_stamp, datetime.datetime)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#

import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from .no_type_key import tNoTypeKey

class tNoType(mettle.io.ISerializable, mettle.db.IPrimaryKey):

    def __init__(self,
                 id: str = '',
                 status: str = '',
                 descr: str = '',
                 grp: str = '',
                 priority: int = 0,
                 pref_corr: dict = None,
                 corr_cnt: int = 0,
                 max_life: int = 0,
                 modified_by: str = ''):
        """
        Constructor.

        :param id: str
        :param status: str
        :param descr: str
        :param grp: str
        :param priority: int
        :param pref_corr: dict
        :param corr_cnt: int
        :param max_life: int
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        """
        self.id = id
        self.status = status
        self.descr = descr
        self.grp = grp
        self.priority = priority
        self.pref_corr = pref_corr
        self.corr_cnt = corr_cnt
        self.max_life = max_life
        self.modified_by = modified_by
        self.tm_stamp = None

    def clear(self):
        """
        Clears all member variables for this database record.
        """
        self.id = ''
        self.status = ''
        self.descr = ''
        self.grp = ''
        self.priority = 0
        self.pref_corr = None
        self.corr_cnt = 0
        self.max_life = 0
        self.modified_by = ''
        self.tm_stamp = None

    def _name(self) -> str:
        """
        Name of the record.

        :return: The name.
        """
        return 'tNoType'

    def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
        """
        Serialize record to a stream.

        :param _w:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _w.write_start(_oname)
        _w.write_string("id", self.id)
        _w.write_char("status", self.status)
        _w.write_string("descr", self.descr)
        _w.write_string("grp", self.grp)
        _w.write_int32("priority", self.priority)
        _w.write_json("pref_corr", self.pref_corr)
        _w.write_int32("corr_cnt", self.corr_cnt)
        _w.write_int32("max_life", self.max_life)
        _w.write_string("modified_by", self.modified_by)
        _w.write_datetime("tm_stamp", self.tm_stamp)
        _w.write_end(_oname)

    def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
        """
        Deserialize record from a stream.

        :param _r:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _r.read_start(_oname)
        self.id = _r.read_string("id")
        self.status = _r.read_char("status")
        self.descr = _r.read_string("descr")
        self.grp = _r.read_string("grp")
        self.priority = _r.read_int32("priority")
        self.pref_corr = _r.read_json("pref_corr")
        self.corr_cnt = _r.read_int32("corr_cnt")
        self.max_life = _r.read_int32("max_life")
        self.modified_by = _r.read_string("modified_by")
        self.tm_stamp = _r.read_datetime("tm_stamp")
        _r.read_end(_oname)

    def _copy_from(self, rec: "tNoType"):
        """
        Copies the column from the rec into this record.

        :param rec: The source record.
        """
        self.id = rec.id
        self.status = rec.status
        self.descr = rec.descr
        self.grp = rec.grp
        self.priority = rec.priority
        self.pref_corr = rec.pref_corr
        self.corr_cnt = rec.corr_cnt
        self.max_life = rec.max_life
        self.modified_by = rec.modified_by
        self.tm_stamp = rec.tm_stamp

    def _write_pk(self, _key: "tNoType"):
        """
        Write the primary key columns to the key object.

        :param _key:
        """
        self._write_key(_key)

    def _read_pk(self, _key: "tNoType"):
        """
        Read the primary key columns from the key object.

        :param _key:
        """
        self._read_key(_key)

    def _serialize_key(self, _w: mettle.io.IWriter):
        """
        Serialize the primary key to a stream.

        :param _w:
        """
        _key = tNoTypeKey()

        self._write_key(_key)
        _key._serialize(_w)

    def _deserialize_key(self, _r: mettle.io.IReader):
        """
        Deserialize the primary key from a stream.

        :param _r:
        """
        _key = tNoTypeKey()

        _key._deserialize(_r)
        self._read_key(_key)

    @staticmethod
    def _cache_davs(dvc=None, for_ins: bool = False, key_only: bool = False):
        """
        Cache the DAV into the dav cache (or create one) and return it.

        :param dvc: (mettle.lib.DavCache), target cache, if None a new one is created and returned.
        :param for_ins: Davs for insert.
        :param key_only: Davs by key only, ie for delete/selectone.
        """
        if dvc is None:
            dvc = mettle.lib.DavCache()

        dvc.add_targ("id", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("id", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 128))
        if key_only:
            return dvc

        dvc.add_targ("status", mettle.lib.Dav(mettle.lib.Dav.eDavType.Min, 1))
        dvc.add_targ("status", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 1))
        dvc.add_targ("status", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("status", mettle.lib.Dav(mettle.lib.Dav.eDavType.In, ['A','D']))
        dvc.add_targ("descr", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("descr", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 256))
        dvc.add_targ("grp", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("grp", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 64))
        dvc.add_targ("priority", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("corr_cnt", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("max_life", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("modified_by", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("modified_by", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 128))

        return dvc

    def _get_davs(self, dvc=None, for_ins: bool = False, key_only: bool = False):
        tNoType._cache_davs(dvc, for_ins, key_only)

    def _write_key(self, _key: tNoTypeKey):
        """
        Write primary key columns to _key.

        :param _key:
        """
        _key.id = self.id

    def _read_key(self, _key: tNoTypeKey):
        """
        Read the primary key columns from _key.

        :param _key:
        """
        self.id = _key.id

    class Status_Couplet(dict):
        key_active = "A"
        key_disabled = "D"

        value_active = "Active"
        value_disabled = "Disabled"

        def __init__(self):
            self[self.key_active] = self.value_active
            self[self.key_disabled] = self.value_disabled

        @staticmethod
        def get_value(key):
            if key == tNoType.Status_Couplet.key_active:
                return tNoType.Status_Couplet.value_active

            if key == tNoType.Status_Couplet.key_disabled:
                return tNoType.Status_Couplet.value_disabled

            return ""

    @staticmethod
    def _pk_compare(x, y) -> int:
        """
        Compare function do primary key comparisons for list sorts and searches.

        :return: zero for x == y, negative for x < y, positive for x > y
        """
        return (x.id > y.id) - (x.id < y.id)

    def __lt__(self, other):
        return other and self._pk_compare(self, other) < 0

    def __gt__(self, other):
        return other and self._pk_compare(self, other) > 0

    def __eq__(self, other):
        return other and self._pk_compare(self, other) == 0

    def __le__(self, other):
        return other and self._pk_compare(self, other) <= 0

    def __ge__(self, other):
        return other and self._pk_compare(self, other) >= 0

    def __ne__(self, other):
        return other == None or self._pk_compare(self, other)

    def __repr__(self) -> str:
        return '<NoType [id:%.32s, status:%.32s, descr:%.32s, grp:%.32s, priority:%r, pref_corr:%.32s, corr_cnt:%r, max_life:%r, modified_by:%.32s, tm_stamp:%s]>' % (
            self.id, self.status, self.descr, self.grp, self.priority, self.pref_corr, self.corr_cnt, self.max_life, self.modified_by, self.tm_stamp)


    class List(list, mettle.io.ISerializable):
        """
        List Class
        """

        def _name(self) -> str:
            """
            Name of the list.

            :return: The name
            """
            return 'tNoType.List'

        def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
            """
            Serialize the list to a stream.

            :param _w:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _w.write_start_list(_oname, len(self))

            for _rec in self:
                _rec._serialize(_w)

            _w.write_end(_oname)

        def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
            """
            Deserialize the list from a stream.

            :param _r:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _cnt = _r.read_start_list(_oname)

            while _cnt >= 1:
                _rec  = tNoType()
                _cnt -= 1
                _rec._deserialize(_r)
                self.append(_rec)

            _r.read_end(_oname)

        def __repr__(self) -> str:
            return '<tNoType.List[%d]>' % (len(self))

        def _get_davs(self, dvc=None, for_ins: bool = False, key_only: bool = False):
            return tNoType._cache_davs(dvc, for_ins, key_only)




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import cerberus

from bs_loco.xloco import xLoco

from bs_loco.db.tables import tCorrType


class BaseLocoTrig:
    """
    This is the base class for firing of loco message requests
    into the system.
    """
    SCHEMA_ADDR = {
        'sms' : {
            'type'      : 'list',
            'required'  : False,
            'empty'     : True,
            'nullable'  : True,
            'schema'    : {
                'type'      : 'string',
                'empty'     : False,
                'nullable'  : False,
                'maxlength' : 64,
            },
        },

        'email' : {
            'type'      : 'list',
            'required'  : False,
            'empty'     : True,
            'nullable'  : True,
            'schema'    : {
                'type'      : 'string',
                'empty'     : False,
                'nullable'  : False,
                'maxlength' : 256,
                'regex'     : '^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$',
            },
        },
    }

    SCHEMA_RESOURCE = {
        'resources' : {
            'type'      : 'list',
            'required'  : True,
            'empty'     : False,
            'nullable'  : False,
            'schema'    : {
                'type'      : 'dict',
                'schema'    : {
                    'filename' : {
                        'type'      : 'string',
                        'empty'     : False,
                        'nullable'  : False,
                        'maxlength' : 256,
                        'required'  : True,
                    },
                    'mime' : {
                        'type'      : 'string',
                        'empty'     : False,
                        'nullable'  : False,
                        'maxlength' : 128,
                        'required'  : False,
                    },
                    'content' : {
                        # 'anyof_type' : [ 'string', 'bytes' ],  #... for some reason bytes is not known
                        'empty'      : False,
                        'nullable'   : False,
                        'required'   : True,
                    },
                    'for' : {
                        'type'      : 'list',
                        'required'  : True,
                        'empty'     : False,
                        'nullable'  : False,
                        'schema'    : {
                            'type'      : 'string',
                            'empty'     : False,
                            'nullable'  : False,
                            'maxlength' : 16,
                            'allowed'   : list(tCorrType.Id_Couplet().keys()),
                        },
                    },
                },
            },
        },
    }


    def __init__(self):
        """
        Constructor.
        """
        self._pod = None
        self._cfg = None


    def id(self) -> str:
        """
        Pure virtual method, the identifier of the content management overload.

        :return: The overload name.
        """
        raise Exception('id() not implemented.')


    def name(self) -> str:
        """
        Pure virtual method, the name of the content management overload.

        :return: The overload name.
        """
        raise Exception('name() not implemented.')


    def is_deadbeef(self) -> bool:
        """
        Virtual method to check if the interface is active but not configured.

        :return: True if this interface is active but not configure.
        """
        return False


    def _validate_dict(self, schema: dict, inp_data: dict, inp_name: str, req: bool) -> dict:
        """
        Ensure the json data matches the input schema.

        :param schema: The cerberus json schema dictionary to check against.
        :param inp_data: The json data to check.
        :param inp_name: The name of the input argument.
        :param req: Indicate if the input is required.
        :return: The inp_data object for convenience.
        """
        if not inp_data:
            if req:
                raise xLoco(f'Invalid arg [{inp_name}] - May not be null or empty')

            return inp_data

        val = cerberus.Validator(schema)

        if not val.validate(inp_data):
            raise xLoco(f'Invalid arg [{inp_name}] - {val.errors}')

        return inp_data



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Tue Mar 30 15:41:12 2021
#  Target Database = postgresql
#
import asyncio
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.msg_hist import tMsgHist
from bs_loco.db.tables.msg_hist_key import tMsgHistKey

from .msg_hist_select_one import dMsgHistSelectOne
from .msg_hist_update import dMsgHistUpdate
from .msg_hist_insert import dMsgHistInsert

class dMsgHist:

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self.rec    = tMsgHist()
        self._dbcon = dbcon

    def __enter__(self):
        """
        DAO enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        DAO exit.
        """
        pass

    def try_select_one(self) -> bool:
        """
        Attempt to select the record.

        :return: True if the record was found.
        """
        _key = tMsgHistKey()

        self.rec._write_key(_key)

        return self.try_select_one_by_key(_key)

    def try_select_one_with(self, _rec: tMsgHist) -> bool:
        """
        Attempt to select the record using a table rec.

        :param _rec: The table rec to select with and into.
        :return: True if the row as selectd.
        """
        _key = tMsgHistKey()
        _rec._write_key(_key)

        if not self.try_select_one_by_key(_key):
            return False

        _rec._copy_from(self.rec)

        return True

    def try_select_one_deft(self,
                            id: int) -> bool:
        """
        Attempt to select the record by it's primary key columns.

        :param id: int
        :return: True if the row as selectd.
        """
        _key = tMsgHistKey()

        _key.id = id

        return self.try_select_one_by_key(_key)

    def try_select_one_by_key(self, _key: tMsgHistKey) -> bool:
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :returns: True if row was selected.
        """
        with dMsgHistSelectOne(self._dbcon) as _qry:
            _qry.exec(_key)
            if not _qry.fetch():
                return False

            self.rec = _qry.orec

        return True

    def select_one(self) -> "dMsgHist":
        """
        Selects the record, raises and error if it does not exist.

        :return: Self for convenience.
        """
        if not self.try_select_one():
            raise mettle.lib.xMettle("SelectOne failed to return a record (dMsgHist)")

        return self

    def select_one_with(self, _rec: tMsgHist) -> "dMsgHist":
        """
        Attempt to select the record with the table record.

        :param _rec: Table record.
        :returns: Self for convenience.
        """
        if not self.try_select_one_with(_rec):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dMsgHist)")

        return self

    def select_one_deft(self,
                        id: int) -> "dMsgHist":
        """
        Selects the record by its primary key columns, raises and error if it does not exist.

        :param id: int
        """
        _key = tMsgHistKey()

        _key.id = id

        if not self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dMsgHist)")

        return self

    def select_one_by_key(self, _key: tMsgHistKey) -> "dMsgHist":
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :return: Self for convenience.
        """
        if not self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dMsgHist)")

        return self

    def update_deft(self,
                    id: int,
                    status: str,
                    msgdate: datetime.datetime,
                    msg: dict):
        """
        Update the record by setting all column values.

        :param id: int
        :param status: str
        :param msgdate: datetime.datetime
        :param msg: dict
        """
        self.rec.id = id
        self.rec.status = status
        self.rec.msgdate = msgdate
        self.rec.msg = msg

        self.update()

    def update(self, irec: tMsgHist = None):
        """
        Update the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to update the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dMsgHistUpdate(self._dbcon) as _qry:
            _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)

    def insert_deft(self,
                    id: int,
                    status: str,
                    msgdate: datetime.datetime,
                    msg: dict):
        """
        Insert the record setting all column values.

        :param id: int
        :param status: str
        :param msgdate: datetime.datetime
        :param msg: dict
        """
        self.rec.id = id
        self.rec.status = status
        self.rec.msgdate = msgdate
        self.rec.msg = msg

        self.insert()

    def insert(self, irec: tMsgHist = None):
        """
        Insert the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to insert into the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dMsgHistInsert(self._dbcon) as _qry:
            _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)

